/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse für Aufgabe, bei der ein Schaubild zusammengeklickt werden soll
*/
function SchaubildTask(
    nodeParent,     // DOM-Knoten, als dessen einzige Kinder das Spiel angezeigt werden soll
    objData         // Spieldaten
) {
    SchaubildTask.parent.call(this);
    this.data = objData;

    if (nodeParent.firstChild)
        nodeParent.innerHTML = '';
    this.msgPlaceGeneral = new ScoTools.MessagePlace(true);
    this.msgPlaceGeneral.appendAsChild(nodeParent);
    this.svg = ScoTools.createHtml('p', nodeParent, null, 'Schaubild wird geladen…');
    this.msgPlace = new ScoTools.MessagePlace();
    this.msgPlace.appendAsChild(nodeParent);

    this.connect();
    this.establish();
    AudioFeedbackTask.cache();
    this.ready = true;
    this.countErrors = Math.max(0, +this.read('suspendData'));
    if (!Number.isFinite(this.countErrors))
        this.countErrors = 0;
    this._loadSvg();
}
SchaubildTask.prototype = Object.create(AudioFeedbackTask.prototype);
SchaubildTask.prototype.constructor = SchaubildTask;
SchaubildTask.parent = AudioFeedbackTask;

// DOM-Repräsentation des SVGs
SchaubildTask.prototype.svg = null;

// Auswahlliste für Textbausteine
SchaubildTask.prototype.liste = null;

// Alles vollständig geladen?
SchaubildTask.prototype.ready = false;

// Informationen zu den Einzelschritten
SchaubildTask.prototype._steps = [];

// Reihenfolge der Auswahlmöglichkeiten und Informationen über deren Sichtbarkeit
SchaubildTask.prototype._itemOrder = [];

// Anzahl der bisher gemachten Fehler
SchaubildTask.prototype.countErrors = 0;

/**
* Die verwendete SVG-Datei laden
* @return undefined
*/
SchaubildTask.prototype._loadSvg = function() {
    var self = this;
    var oReq = new XMLHttpRequest();
    oReq.overrideMimeType('text/xml');
    oReq.addEventListener("load", function() {
        self._importSvg(oReq.responseXML);
    });
    oReq.addEventListener("error", function() {
        self.showMsg('Das Schaubild kann wegen eines unerwarteten Fehlers leider nicht geladen werden…', false, true);
    });
    oReq.addEventListener("abort", function() {
        self.showMsg('Das Schaubild kann nicht geladen werden, weil du den Ladevorgang abgebrochen hast.', false, true);
    });
    oReq.open('GET', 'schaubild.svg');
    oReq.send();
}

/**
* Array mit direkten Kindknoten erstellen, die angegebenen Tagnamen haben
* @return Array
*/
SchaubildTask.prototype._getDirectChildrenByTagName = function(
    parent,     // Elternknoten
    tagname     // gewünschter Tagname in Großbuchstaben
) {
    var a = [];
    for (var node = parent.firstChild; node; node = node.nextSibling) {
        if ((node.nodeType == Node.ELEMENT_NODE) && (node.nodeName.toUpperCase() == tagname))
            a.push(node);
    }
    return a;
}

/**
* Den nächsten erlaubten Index aus this.data.parts auslesen
* @return Array     Feld mit dem oder den nächsten Indizes
*/
SchaubildTask.prototype._getNextIndizes = function() {
    if (!this._lastIndex)
        this._lastIndex = -Infinity;
    var self = this;
    var thisIndex = Infinity;
    var indizes;
    var result = [];
    for (var n = 0; n < this.data.parts.length; ++n) {
        if (this.data.parts[n].index === undefined)
            continue;
        indizes = this.data.parts[n].index;
        if (!(indizes instanceof Array))
            indizes = [ indizes ];
        indizes = indizes.filter(function(v) { return ((v > self._lastIndex) && (v <= thisIndex)); });
        if (!indizes.length)
            continue;
        indizes.sort(function(a, b) { return a - b; });
        if (indizes[0] < thisIndex) {
            thisIndex = indizes[0];
            result = [ n ];
        } else if (indizes[0] == thisIndex)
            result.push(n);
    }
    this._lastIndex = thisIndex;
    return result;
}

/**
* Schaubild aus SVGDocument importieren, und das - ebenso wie die Auswahlliste - anzeigen
* @return undefined
*/
SchaubildTask.prototype._importSvg = function(
    xml         // Document mit SVG
) {
    var self = this;
    var level = (this.isConnected() && +this.read('lessonLocation')) || 0;
    var svg = xml.documentElement;
    svg.removeAttribute('class');
    svg = document.importNode(svg, true);
    this.ready = true;

    var z;
    var arrG = this._getDirectChildrenByTagName(svg, 'G');
    for (z = 1; z < arrG.length; ++z) {
        this._steps.push({
            svg: arrG[z],
            correct: this._getNextIndizes(),
            virtualIndex: this._lastIndex
        });
        if (level < z)
            arrG[z].parentNode.removeChild(arrG[z]);
    }
    //while (z = this._getNextIndizes())
        //this._steps.push({ svg: null, correct: z });

    this.svg.parentNode.replaceChild(svg, this.svg);
    this.svg = svg;
    this.liste = new Auswahlliste();
    this.liste.onclick = function(id) {
        self.procClick(id);
    };
    this.svg.parentNode.appendChild(this.liste.node);
    this._updateList();
    this._showMsg();
}

// In welcher Stufe des Bildaufbaus befinden wir uns?
Object.defineProperty(SchaubildTask.prototype, 'level', {
    enumerable: true,
    get() {
        var r = 0;
        while ((r < this._steps.length) && this._steps[r].svg && this._steps[r].svg.parentNode)
            ++r;
        return r;
    }
});

/**
* Die passende Meldung anzeigen (Kurzerklärung bzw. Erfolgsmeldung)
* @return undefined
*/
SchaubildTask.prototype._showMsg = function() {
    if (this.level >= this._steps.length) {
        var text = '';
        if (this.data.done)
            text = this.data.done;
        else if (this.countErrors == 0)
            text = 'Du hast diese Aufgabe fehlerfrei gelöst!';
        else if (this.countErrors == 1)
            text = 'Du hast diese Aufgabe gelöst mit einem Fehlklick';
        else
            text = 'Du hast diese Aufgabe nach ' + this.countErrors + ' Fehlklicks gelöst';
        this.msgPlace.add(text, true);
    } else {
        this.msgPlace.add(this.data.task ? this.data.task : 'Ergänze die fehlenden Schritte, indem du sie in der richtigen Reihenfolge anklickst', ScoTools.MSGTYPE_INFO);
    }
}

/**
* Sicherstellen, dass die richtigen Einträge in der Liste stehen
* @return undefined
*/
SchaubildTask.prototype._updateList = function() {
    var level = this.level;
    if (level >= this._steps.length) {
        this.liste.clear();
        this.liste.add('Von vorne starten', Infinity);
        this.liste.setClass(Infinity, 'reload');
        this._showMsg();
        return;
    }
    if (this._steps[level].correct.length) {
        if (level && this._steps[level - 1].correct.length && this.liste.getLength())
            return;
        var liste = this.liste;
        liste.clear();
        var parts = this.data.parts;//.filter(({ hidden }, index, arr) => (!(hidden instanceof Array) || (this._steps[level].correct.indexOf(index) != -1) || (hidden.indexOf(this._steps[level].virtualIndex) == -1)));
        if (this.data.shuffle)
            parts = parts.map(function(o, id) { return { 'text': o.text, 'hidden': o.hidden, 'id': id, 'rand': Math.random() }; }).sort(function(a, b) { return a.rand - b.rand; })
        else
            parts = parts.map(function(o, id) { return { 'text': o.text, 'hidden': o.hidden, 'id': id }; });
        parts.forEach(function(part) { liste.add(part.text, part.id); });
        this._itemOrder = parts.map(({ id, text, hidden }) => ({ id, text, hidden, visual: true }));
        this.showAndHideItems(-Infinity);
    } else {
        if (level && !this._steps[level - 1].correct.length && this.liste.getLength())
            return;
        this.liste.clear();
        this.liste.add('Weiter', -Infinity);
        this.liste.setClass(-Infinity, 'forward');
    }
}

/**
* Auswahlmöglichkeiten abhängig vom aktuellen Schritt anzeigen oder verbergen
* @return undefined
*/
SchaubildTask.prototype.showAndHideItems = function(
    lastIndex       // Optional: Interner Index der zuletzt angeklickten Auswahlmöglichkeit, die garantiert sichtbar bleiben soll
) {
    const nextIndex = this._steps[Math.min(this.level, this._steps.length - 1)].virtualIndex;
    this._itemOrder.forEach(itemInfo => {
        const doShow = ((itemInfo.id == lastIndex) || !(itemInfo.hidden instanceof Array) || (itemInfo.hidden.indexOf(nextIndex) == -1));
        if (doShow == itemInfo.visual)
            return;
        if (doShow)
            this.liste.add(itemInfo.text, itemInfo.id);
        else
            this.liste.remove(itemInfo.id);
        itemInfo.visual = doShow;
    });
}

/**
* Darauf reagieren, wenn ein Listeneintrag angeklickt wird
* @return undefined
*/
SchaubildTask.prototype.procClick = function(
    id      // Interne ID
) {
    if (id == Infinity) {
        this.reset();
        return;
    }
    if (this.canSave() && (this.read('lessonStatus') == 'not attempted'))
        this.write('lessonStatus', 'incomplete');
    var level = this.level;
    if ((id == -Infinity) || (this._steps[level].correct && (this._steps[level].correct.indexOf(id) != -1))) {
        this.liste.setOnlyClass(id, 'success');
        if (this.canSave())
            this.write('lessonLocation', level + 1);
        this.next(id);
        this.playAudio(true);
        return;
    }
    this.liste.setOnlyClass(id, 'fail');
    ++this.countErrors;
    if (this.canSave())
        this.write('suspendData', this.countErrors);
    this.playAudio();
}

/**
* Das Schaubild komplett zurücksetzen
* @return undefined
*/
SchaubildTask.prototype.reset = function() {
    this._steps.map(function(step) {
        if (step.svg.parentNode)
            step.svg.parentNode.removeChild(step.svg);
    });
    this._updateList();
    this._showMsg();
}

/**
* Das Schaubild um einen weiteren Schritt vervollständigen
* @return undefined
*/
SchaubildTask.prototype.next = function(
    lastIndex       // Optional: Interner Index der zuletzt angeklickten Auswahlmöglichkeit, die garantiert sichtbar bleiben soll
) {
    var level = this.level;
    this.svg.appendChild(this._steps[level].svg);
    if (++level >= this._steps.length) {
        this.saveScore(Math.max(0, 100 - this.countErrors), 0, 100, 'completed');
        this.showFinalMessage();
    }
    this._updateList();
    this.showAndHideItems(lastIndex);
}
